// (c) Peter Kankowski, 2006 http://smallcode.weblogs.us
#include <windows.h>

#define NUM_OF(A) (sizeof(A)/sizeof((A)[0])) // Number of elements in array

void printf(const char* spec, ...) {
	CHAR buff[1024];
	va_list va;
	va_start(va, spec);
	DWORD len = wvsprintfA(buff, spec, va);
	WriteFile(GetStdHandle(STD_OUTPUT_HANDLE), buff, len, &len, NULL);
}

// Faster strlen function. Warning: in rare cases it may
// cause page faults (if the string is at the end of
// the page). To avoid this, you must align the start
// of the string by 4 bytes, or you must add 3 extra
// bytes to the end of the string.
// E.g., instead of char* str = (char*)malloc(size); use
//                  char* str = (char*)malloc(size + 3);
int strlen_my(const char *s) {
	int len = 0;
	for(;;) {
		unsigned x = *(unsigned*)s;
		if((x & 0xFF) == 0) return len;
		if((x & 0xFF00) == 0) return len + 1;
		if((x & 0xFF0000) == 0) return len + 2;
		if((x & 0xFF000000) == 0) return len + 3;
		s += 4, len += 4;
	}
}

int strlen_std(const char *s) { // Wrapper around strlen
	return (int)strlen(s);
}

int strlen_AgnerFog(const char* s) {
	__asm {
        mov     eax, [s]               ; get pointer s 
        lea     edx, [eax+3]           ; pointer+3 used in the end 
l1:     
		mov     ebx, [eax]             ; read 4 bytes of string 
        add     eax, 4                 ; increment pointer 
        lea     ecx, [ebx-01010101H]   ; subtract 1 from each byte 
        not     ebx                    ; invert all bytes 
        and     ecx, ebx               ; and these two 
        and     ecx, 80808080H         ; test all sign bits 
        jz      l1                     ; no zero bytes, continue loop 

        mov     ebx, ecx 
        shr     ebx, 16 
        test    ecx, 00008080H         ; test first two bytes 
        cmovz   ecx, ebx               ; shift if not in first 2 bytes 
        lea     ebx, [eax+2]           ; .. and increment pointer by 2 
        cmovz   eax, ebx 
        add     cl,  cl                ; test first byte 
        sbb     eax, edx               ; compute length 
	}
}

// Test data
char *txt[] = {"Four","score","and","seven","years","ago","our",
 "fathers","brought","forth","upon","this","continent","a","new","nation",
 "conceived","in","liberty","and","dedicated","to","the","proposition",
 "that","all","men","are","created","equal"};

char jabberwocky[] =
"'Twas brillig, and the slithy toves\n"
"Did gyre and gimble in the wabe:\n"
"All mimsy were the borogoves,\n"
"And the mome raths outgrabe.\n"

"Beware the Jabberwock, my son!\n"
"The jaws that bite, the claws that catch!\n"
"Beware the Jubjub bird, and shun\n"
"The frumious Bandersnatch!\n"

"He took his vorpal sword in hand:\n"
"Long time the manxome foe he sought -\n"
"So rested he by the Tumturn tree,\n"
"And stood awhile in thought.\n"

"And, as in uffish thought he stood;\n"
"The Jabberwock, with eyes of flame,\n"
"Came whiffling through the tulgey wood,\n"
"And burbled as it came!\n"

"One, two! One, two! And through and through\n"
"The vorpal blade went snicker-snackl\n"
"He left it dead, and with its head\n"
"He went galumphing back.\n"

"And hast thou slain the Jabberwock?\n"
"Come to my arms, my beamish boy!\n"
"O frabjous day! Callooh! Callay!\n"
"He chortled in his joy.\n"

"'Twas brillig, and the slithy toves\n"
"Did gyre and gimble in the wabe:\n"
"All mimsy were the borogoves,\n"
"And the mome raths outgrabe.";


unsigned __int64 GetRDTSC() {
	__asm {
		; Flush the pipeline
		XOR eax, eax
		CPUID
		; Get RDTSC counter in edx:eax
		RDTSC
	}
}

typedef int(*STRLEN)(const char *s);
typedef UINT(*TESTFUNC)(STRLEN strlenfunc);

UINT __forceinline Test1(STRLEN strlenfunc) {
	UINT a = 0;
	for(int i = 0; i < NUM_OF(txt); ++i)
		a += strlenfunc(txt[i]);
	return a;
}

UINT __forceinline Test2(STRLEN strlenfunc) {
	return strlenfunc(jabberwocky);
}

// Profiling function
template <TESTFUNC test>
void TestStrLen(const char* label, STRLEN strlenfunc) {
	for(int j = 0; j < 20; ++j) {
		SetPriorityClass(GetCurrentProcess(), REALTIME_PRIORITY_CLASS);
		SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_TIME_CRITICAL);
		UINT64 start = GetRDTSC();
		UINT a = test(strlenfunc);
		UINT64 time = GetRDTSC() - start;
		SetPriorityClass(GetCurrentProcess(), NORMAL_PRIORITY_CLASS);
		SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL);
		printf("%10s: %10I64d ticks, a = %d\n", label, time, a);
	}
}

int main() {
	printf("strlen - short strings\n");
	TestStrLen<Test1>("Microsoft", strlen_std);
	TestStrLen<Test1>("Peter", strlen_my);
	TestStrLen<Test1>("A.Fog", strlen_AgnerFog);
	printf("\nstrlen - long string\n");
	TestStrLen<Test2>("Microsoft", strlen_std);
	TestStrLen<Test2>("Peter", strlen_my);
	TestStrLen<Test2>("A.Fog", strlen_AgnerFog);
	return 0;
};